#------------------------------------------------------------------
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/Siebel/Siebel.pm,v 1.27 2011/08/05 10:26:18 sgadgil Exp $
#
# $Author: sgadgil $
# $Date: 2011/08/05 10:26:18 $
# $Revision: 1.27 $
#
# Provides:                            Siebel
# Author:                              Anshuman Atre
# Creation Date:                       04 Jul, 2006
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package Siebel;

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use File::Path qw ( mkpath );
use File::stat;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSDefines;     # This contains constants that VCS uses
use VCSagentFW;     # Primary Agent Interface to the ACC Library

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

my $sCmdHares  = $HARES || "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE || "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = $HADECRYPT || "$CLUS_HOME/sbin/${sPrefix}decrypt";

my @liRunningProcs = (); # Global..
my $bSharedLogin = 0; # Default: Unique..

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: Siebel.pm,v 1.27 2011/08/05 10:26:18 sgadgil Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define Siebel subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String EnvFile,
                              IN String HostName,
                              IN String SiebelRoot,
                              IN Integer SiebelUser,
                              IN String Sadmin,
                              IN String SadminCrPasswd,
                              IN String SiebelGWHost,
                              IN Integer SiebelGWPort,
                              IN String ServerName,
                              IN String ServerType,
                              IN String SiebelEnterprise,
                              IN String CompGrps,
                              IN String MonitorProgram,
                              IN Integer SecondLevelMonitor );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localEnvFile                  Absolute path to enviroment file for Siebel Server
 $localHostName                 Hostname/IPv4 address of server hosting this server
 $localSiebelRoot               Absolute path to location of Siebel Server
 $localSiebelUser               UNIX user for Siebel Server
 $localSadmin                   Siebel level administrative login
 $localSadminCrPasswd           Crypted passwd for Siebel administrative login
 $localSiebelGWHost             Virtual Hostname hosting the Gateway
 $localSiebelGWPort             Port allocated for the Gateway
 $localServerName               Siebel Server name if type is SRVR
 $localServerType               Type of Siebel Server
 $localSiebelEnterprise         Siebel Enterprise of this Server
 $localCompGrps                 Comma separated list of compgrps to be monitored
 $localMonitorProgram           Full path to external monitor program
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localEnvFile,
       $localHostName,
       $localSiebelRoot,
       $localSiebelUser,
       $localSadmin,
       $localSadminCrPasswd,
       $localSiebelGWHost,
       $localSiebelGWPort,
       $localServerName,
       $localServerType,
       $localSiebelEnterprise,
       $localCompGrps,
       $localMonitorProgram,
       $localSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  SiebelRoot               [%s]
  SiebelUser               [%s]
  Sadmin                   [%s]
  SadminCrPasswd           [%s]
  SiebelGWHost             [%s]
  SiebelGWPort             [%s]
  ServerName               [%s]
  ServerType               [%s]
  SiebelEnterprise         [%s]
  CompGrps                 [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,   $sEntryPointName,   $localEnvFile,    $localHostName,
  $localSiebelRoot,         $localSiebelUser,         $localSadmin,
  $localSadminCrPasswd,   $localSiebelGWHost,   $localSiebelGWPort,
  $localServerName,    $localServerType,    $localSiebelEnterprise,
  $localCompGrps, $localMonitorProgram, $localSecondLevelMonitor );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #------------------------------------------
  # Get relevant references out of arg list..
  #------------------------------------------
  my $rsSiebelUser = \$_[4]; # Holds localSiebelUser..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName    EnvFile    HostName  SiebelRoot      SiebelUser
       Sadmin  SadminCrPasswd   SiebelGWHost   SiebelGWPort   ServerName
       ServerType     SiebelEnterprise      CompGrps      MonitorProgram
       SecondLevelMonitor );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( HostName SiebelUser ServerType EnvFile SiebelRoot
                           SiebelGWHost SiebelEnterprise );
  my @lsOfflineArgs = qw ( SiebelUser ServerType EnvFile SiebelRoot SiebelEnterprise );
  my @lsMonitorArgs = qw ( HostName SiebelUser ServerType SiebelRoot SiebelEnterprise );
  my @lsCleanArgs   = qw ( SiebelUser ServerType EnvFile SiebelRoot SiebelEnterprise );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-------------------------------
  # Individual validation starts..
  #-------------------------------

  #---------------------------
  # Check if login is shared..
  #---------------------------
  if ( $localSiebelUser =~ s/^\+// ) {
    ++ $bSharedLogin;                  # Set global..
    $$rsSiebelUser = $localSiebelUser; # Update..
  }

  #-----------------------
  # Validate Siebel user..
  #-----------------------
  unless ( defined getpwnam $localSiebelUser ) {
    &logError ( 20003, "<%s> SiebelUser [%s] not a valid UNIX login",
    $sMyName, $localSiebelUser );
    $iReturnCode = 0; # Failure..
  }

  #----------------------
  # Validate ServerType..
  #----------------------
  if ( $localServerType !~ /^\b(SRVR|GTWY)\b$/o ) {
    &logError ( 20004,
    "<%s> ServerType [%s] has to be one of SRVR or GTWY",
    $sMyName, $localServerType );
    $iReturnCode = 0; # Failure..
  }

  #----------------------
  # Validate ServerName..
  #----------------------
  if ( $localServerType eq 'SRVR' &&
       $localServerName !~ /\S/o ) {
    &logError ( 20005,
    "<%s> ServerName cannot be NULL for ServerType [%s]",
    $sMyName, $localServerType );
    $iReturnCode = 0; # Failure..
  }

  #--------------------------------------------
  # Validate virtual hostname for this server..
  #--------------------------------------------
  if ( $sEntryPointName =~ /^(online|monitor)$/o ) {
    #--------------------
    # Validate hostname..
    #--------------------
    if ( ! inet_aton ( $localHostName ) ) {
      &logError ( 20006, "<%s> HostName [%s] is invalid",
      $sMyName, $localHostName );
      $iReturnCode = 0; # Failure..
    }

    #---------------------------------------------------
    # For GTWY HostName and SiebelGWHost are identical..
    #---------------------------------------------------
    if ( $localServerType eq 'GTWY' &&
         $localHostName ne $localSiebelGWHost ) {
      &logError ( 20007,
      "<%s> HostName [%s] and SiebelGWHost [%s] not identical for ServerType [%s]",
      $sMyName, $localHostName, $localSiebelGWHost, $localServerType );
      $iReturnCode = 0; # Failure..
    }
  }

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-----------------------------
  # Return if clean or monitor..
  #-----------------------------
  return $iReturnCode if ( $sEntryPointName =~ /^(clean|monitor)$/o );

  #----------------------------
  # Validate environment file..
  #----------------------------
  if ( ! -e $localEnvFile ) {
    &logError ( 20008,
    "<%s> EnvFile [%s] does not exist", $sMyName, $localEnvFile );
    $iReturnCode = 0; # Failure..
  }

  #-----------------------------------------
  # Validate Siebel installation directory..
  #-----------------------------------------
  if ( ! -d $localSiebelRoot ) {
    &logError ( 20009, "<%s> Server Root [%s] does not exist",
    $sMyName, $localSiebelRoot );
    $iReturnCode = 0; # Failure..
  }

  #---------------------------
  # Validate for online only..
  #---------------------------
  if ( $sEntryPointName eq 'online' ) {
    #------------------------------------
    # For Siebel Enterprise server only..
    #------------------------------------
    if ( $localServerType eq 'SRVR' && # Siebel Enterprise Server..
         $localSecondLevelMonitor   && # with SLM enabled..
         $localSadmin !~ /\S/ ) {      # and Sadmin not specified..
      &logError ( 20010,
      "<%s> Sadmin [%s] cannot be NULL for ServerType [%s] and SecondLevelMonitor [%s]",
      $sMyName, $localSadmin, $localServerType, $localSecondLevelMonitor );
      $iReturnCode = 0; # Failure..
    }

    if ( $localSadmin =~ /\S/o &&
         $localSecondLevelMonitor ) {
      if ( $localServerType eq 'SRVR' ) {
        #--------------------
        # Validate CompGrps..
        #--------------------
        if ( $localCompGrps !~ /([^(,|\s)]+)/g ) {
          &logTrace ( 20011,
          "<%s> CompGrps NULL for Sadmin [%s] and SecondLevelMonitor [%s]",
          $sMyName, $localSadmin, $localSecondLevelMonitor );
        }
      }
      else { # Sadmin specified for GTWY..
        &logTrace ( 20012,
        "<%s> Sadmin [%s] specified for ServerType [%s].
        Authentication enforced for Siebel GTWY Server",
        $sMyName, $localSadmin, $localServerType );
      }

      #-----------------------------
      # Validate SadminCryptPasswd..
      #-----------------------------
      if ( $localSadminCrPasswd !~ /\S/o ) {
        &logError ( 20013,
        "<%s> SadminCrPasswd cannot be NULL for Sadmin [%s] and SecondLevelMonitor [%s]",
        $sMyName, $localSadmin, $localSecondLevelMonitor );
        $iReturnCode = 0; # Failure..
      }
    }

    #-------------------------
    # For both SRVR and GTWY..
    #-------------------------
    unless ( Siebel::ValidateSiebelParams ( $localSiebelEnterprise ) ) {
      &logError ( 20014, "<%s> Configuration error in Siebel Enterprise [%s]",
      $sMyName, $localSiebelEnterprise );
    }

    #---------------------------------
    # Validate Siebel Gateway Server.. 
    #---------------------------------
    if ( ! inet_aton ( $localSiebelGWHost ) ) {
      &logError ( 20015, "<%s> SiebelGWHost [%s] is invalid",
      $sMyName, $localSiebelGWHost );
      $iReturnCode = 0; # Failure..
    }

    #--------------------------------------
    # MonitorProgram has to be executable..
    #--------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {                        # And not executable..
      &logError ( 20017,
      "<%s> MonitorProgram [%s] not executable", $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }
  }

  $iReturnCode ||
  &logWarn ( 20018,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String SiebelRoot,
                         IN String SiebelUser,
                         IN String SiebelGWHost,
                         IN Integer SiebelGWPort,
                         IN String EnvFile,
                         IN String ServerName,
                         IN String ServerType,
                         IN String Sadmin,
                         IN String SadminCrPasswd,
                         IN String CompGrps );

 Description : Start logic needed for online entry pt

 Parameters :

 $sSiebelRoot               Absolute path to location of Siebel Server
 $sSiebelUser               UNIX user for Siebel Server
 $sSiebelGWHost             Virtual Hostname hosting the Gateway
 $iSiebelGWPort             Port allocated for the Gateway
 $sEnvFile                  Absolute path to enviroment file for Siebel Server
 $sServerName               Siebel Server name if type is SRVR
 $sServerType               Type of Siebel Server
 $sSiebelEnterprise         Siebel Enterprise of this Server
 $sServerType               Type of Siebel Server
 $sSadmin                   Siebel level administrative login
 $sSadminCrPasswd           Crypted passwd for Siebel administrative login
 $sCompGrps                 Comma separated list of compgrps to be monitored
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelRoot,
       $sSiebelUser,
       $sSiebelGWHost,
       $iSiebelGWPort,
       $sEnvFile,
       $sServerName,
       $sServerType,
       $sSiebelEnterprise,
       $sSadmin,
       $sSadminCrPasswd,
       $sCompGrps,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20019, "Subroutine <%s> called with:
  SiebelRoot               [%s]
  SiebelUser               [%s]
  SiebelGWHost             [%s]
  SiebelGWPort             [%s]
  EnvFile                  [%s]
  ServerName               [%s]
  ServerType               [%s]
  SiebelEnterprise         [%s]
  Sadmin                   [%s]
  SadminCrPasswd           [%s]
  CompGrps                 [%s]",

  $sMyName,     $sSiebelRoot,     $sSiebelUser,     $sSiebelGWHost,
  $iSiebelGWPort,     $sEnvFile,     $sServerName,    $sServerType,
  $sSiebelEnterprise, $sSadmin, $sSadminCrPasswd, $sCompGrps );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + Siebel::GetEngineTimeout ( 1 ); # With margin..

  #-----------------------------------------
  # Whack any preexisting Siebel processes..
  #-----------------------------------------
  &logTrace ( 20020,
  "<%s> Killing processes owned by [%s]", $sMyName, $sSiebelUser );

  if ( Siebel::KillSiebelProcesses ( $sSiebelUser,
                                     $sSiebelRoot,
                                     $sServerName,
                                     $sServerType,
                                     $sSiebelEnterprise ) ) {
    &logTrace ( 20021,
    "<%s> Processes for User [%s] do not exist before startup",
    $sMyName, $sSiebelUser );
  }
  else {
    &logError ( 20022,
    "<%s> Could not kill all processes for User [%s]",
    $sMyName, $sSiebelUser );
    $iReturnCode = 0; # Failure..
  }

  #-----------------------------------
  # Remove preexisting IPC resources..
  #-----------------------------------
  unless ( $bSharedLogin ) {
    if ( Proc::RemoveIPCResourcesForUser ( $sSiebelUser, O_IPCALL ) ) {
      &logTrace ( 20023,
      "<%s> IPC resources for User [%s] do not exist before startup",
      $sMyName, $sSiebelUser );
    }
    else {
      &logError ( 20024,
      "<%s> Could not remove all IPC resources for User [%s]",
      $sMyName, $sSiebelUser );
      $iReturnCode = 0; # Failure..
    }
  }

  #-------------------------------------------
  # Declare start command and args variables..
  #-------------------------------------------
  my ( $sCommand, $sCommandArgs, $sSvcFile, $sOsdfFile, $sShmFile,
       $iPid, $iPCpu, $iTimeout, $iSleepTime, $iExitCode,
       @lsCmdout, @lsFiles, @liPids ) = ();

  #-----------------------------
  # Select correct server type..
  #-----------------------------
  if ( $sServerType eq 'GTWY' ) {
    $sCommand = "$sSiebelRoot/bin/start_ns";
    if ( $iSiebelGWPort > 0 ) {
      $sCommandArgs = "-p $iSiebelGWPort";
    }
    $sSvcFile = "$sSiebelRoot/sys/svc.gtwyns.";

    #--------------------------------
    # Identify corrupt service file..
    #--------------------------------
    if ( -z $sSvcFile ) { # Zero byte file..
      &logWarn ( 20025, "<%s> Located empty service file [%s]",
      $sMyName, $sSvcFile );
      push ( @lsFiles, $sSvcFile );
    }

    #-------------------
    # Remove the files..
    #-------------------
    if ( scalar @lsFiles &&
         ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
      &logError ( 20026, "<%s> Failed to remove stale file(s) [%s]",
      $sMyName, join ( ' ', @lsFiles ) );
      return 0;
    }
  }
  else { # Siebel Server..
    #-------------------------
    # Define files to remove..
    #-------------------------
    $sOsdfFile = "$sSiebelRoot/sys/osdf.$sSiebelEnterprise.$sServerName";
    $sShmFile = "$sSiebelRoot/admin/$sSiebelEnterprise.$sServerName.shm";
    $sSvcFile = "$sSiebelRoot/sys/svc.siebsrvr.$sSiebelEnterprise:$sServerName";

    #-------------------------------------------
    # Identify .shm and .osdf files if present..
    #-------------------------------------------
    if ( -z $sOsdfFile ) {
      &logTrace ( 20027, "<%s> Stale OSDF file [%s] located for removal",
      $sMyName, $sOsdfFile );
      push ( @lsFiles, $sOsdfFile );
    }

    if ( -e $sShmFile ) {
      &logTrace ( 20028, "<%s> Stale SHM file [%s] located for removal",
      $sMyName, $sShmFile );
      push ( @lsFiles, $sShmFile );
    }

    #--------------------------------
    # Identify corrupt service file..
    #--------------------------------
    if ( -z $sSvcFile ) { # Zero byte file..
      &logWarn ( 20029, "<%s> Located empty service file [%s]",
      $sMyName, $sSvcFile );
      push ( @lsFiles, $sSvcFile );
    }

    #-------------------
    # Remove the files..
    #-------------------
    if ( scalar @lsFiles &&
         ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
      &logError ( 20030, "<%s> Failed to remove stale file(s) [%s]",
      $sMyName, join ( ' ', @lsFiles ) );
      $iReturnCode = 0;
    }

    $iReturnCode || return $iReturnCode; # Return, if failed..

    #--------------------------
    # Re-set command and args..
    #--------------------------
    $sCommand = "$sSiebelRoot/bin/start_server";
    $sCommandArgs = "-e $sSiebelEnterprise -g $sSiebelGWHost";
    if ( $iSiebelGWPort > 0 ) {
      $sCommandArgs = $sCommandArgs . ":$iSiebelGWPort ";
    }
    $sCommandArgs = $sCommandArgs . " $sServerName";
  }

  #-----------------------------------
  # Re-create service file if needed..
  #-----------------------------------
  if ( ! -e $sSvcFile ) {
    $iTimeout = $iTillWhen - time; # Get bearings..
    $iReturnCode = Siebel::RecreateSvcFile ( $sSiebelUser,
                                             $sEnvFile,
                                             $sSiebelRoot,
                                             $sSiebelEnterprise,
                                             $sServerName,
                                             $sSiebelGWHost,
                                             $iSiebelGWPort,
                                             $iTimeout );
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  &logTrace ( 20031,
  "<%s> Starting Siebel [%s] with Command [%s] Args [%s]",
  $sMyName, $sServerType, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20032,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          Siebel::ExecLocalCmd ( $sSiebelUser,
                                 $sEnvFile,
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout ) ) {

    ( $iPid ) = $lsCmdout [-1] =~ /(\d+)$/; # Get PID..

    #--------------------------
    # If PID is really online..
    #--------------------------
    if ( kill ( 0, $iPid ) ) {
      &logTrace ( 20033,
      "<%s> Cmd [%s] with args [%s] executed successfully.
      PID of command is [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iPid );
  
      #----------------------------------------------
      # Derive our resource type and resource names..
      #----------------------------------------------
      my $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
      my $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
      my $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";
      my $sPidFile = "$sPIDDir/pidfile.$sServerType";
  
      #------------------
      # Create PID file..
      #------------------
      if ( Siebel::CreatePidFile ( $sPidFile,
                                   $iPid ) ) {
        &logTrace ( 20034,
        "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
      }
      else {
        &logError ( 20035,
        "<%s> Could not create PID file [%s] successfully",
        $sMyName, $sPidFile );
      }
    }
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;
  
  #--------------------
  # Delay entry point..
  #--------------------
  $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20036,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         Siebel::GetProcState ( \@liPids,
                                $sSiebelRoot,
                                $sSiebelUser,
                                $sServerName,
                                $sServerType,
                                $sSiebelEnterprise,
                                1 ) &&
         ( $iSecondLevelMonitor   ? 
         ( $sServerType eq 'SRVR' ?
         ( ( $iPCpu = Siebel::PercentCpuUtil ( $liPids [0] ) ) &&
         ( $iPCpu < 1.0 ) &&
         ( VCSDef_MonitorExit::ONLINE ==              # Online after..
           Siebel::SecondLevelMonitor ( $sSiebelRoot, # Running SLM..
                                        $sSiebelUser,
                                        $sServerName,
                                        $sServerType,
                                        $sSadmin,
                                        $sSadminCrPasswd,
                                        $sSiebelGWHost,
                                        $iSiebelGWPort,
                                        $sEnvFile,
                                        $sSiebelEnterprise,
                                        $sCompGrps,
                                        $iTillWhen ) ) ) : 1 ) : 1 ) ) {

        &logInfo ( 20037,
        "<%s> Confirmed startup of Siebel [%s] with PID [%s]",
        $sMyName, $sServerType, join ( ' ', @liPids ) );

        $iReturnCode = 1; # Success..
        last;
    }
    else {
      &logTrace ( 20038,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0;  # Failure..
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String SiebelRoot,
                        IN String SiebelUser,
                        IN String EnvFile,
                        IN String ServerName,
                        IN String ServerType,
                        IN String SiebelEnterprise );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sSiebelRoot               Absolute path to location of Siebel Server
 $sSiebelUser               UNIX user for Siebel Server
 $sEnvFile                  Absolute path to enviroment file for Siebel Server
 $sServerName               Siebel Server name if type is SRVR
 $sServerType               Type of Siebel Server
 $sSiebelEnterprise         Siebel Enterprise of this Server

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelRoot,
       $sSiebelUser,
       $sEnvFile,
       $sServerName,
       $sServerType,
       $sSiebelEnterprise ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20039, "Subroutine <%s> called with:
  SiebelRoot               [%s]
  SiebelUser               [%s]
  EnvFile                  [%s]
  ServerName               [%s]
  ServerType               [%s]
  SiebelEnterprise         [%s]",

  $sMyName, $sSiebelRoot,  $sSiebelUser,  $sEnvFile,  $sServerName,
  $sServerType, $sSiebelEnterprise );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + Siebel::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsCmdout ) = ();

  #-----------------------------
  # Select appropriate command..
  #-----------------------------
  if ( $sServerType eq 'GTWY' ) {
    $sCommand = "$sSiebelRoot/bin/stop_ns";
    $sCommandArgs = "";
  }
  else { # Siebel Server..
    $sCommand = "$sSiebelRoot/bin/stop_server";
    $sCommandArgs = "-e $sSiebelEnterprise -M $sServerName";
  }

  &logTrace ( 20040,
  "<%s> Stopping Siebel [%s] with Command [%s] Args [%s]",
  $sMyName, $sServerType, $sCommand, $sCommandArgs );

  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20041,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          Siebel::ExecLocalCmd ( $sSiebelUser,
                                 $sEnvFile,
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout ) ) {
    &logTrace ( 20042,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #----------------------------------------
  # Whack any persisting Siebel processes..
  #----------------------------------------
  &logTrace ( 20043,
  "<%s> Killing processes owned by [%s]", $sMyName, $sSiebelUser );

  if ( Siebel::KillSiebelProcesses ( $sSiebelUser,
                                     $sSiebelRoot,
                                     $sServerName,
                                     $sServerType,
                                     $sSiebelEnterprise ) ) {
    &logTrace ( 20044,
    "<%s> Processes for User [%s] do not exist after shutdown",
    $sMyName, $sSiebelUser );
  }
  else {
    &logError ( 20045,
    "<%s> Could not kill all processes for User [%s]",
    $sMyName, $sSiebelUser );
    $iReturnCode = 0; # Failure..
  }

  #----------------------------------
  # Remove persisting IPC resources..
  #----------------------------------
  unless ( $bSharedLogin ) {
    if ( Proc::RemoveIPCResourcesForUser ( $sSiebelUser, O_IPCALL ) ) {
      &logTrace ( 20046,
      "<%s> IPC resources for User [%s] do not exist after shutdown",
      $sMyName, $sSiebelUser );
    }
    else {
      &logError ( 20047,
      "<%s> Could not remove all IPC resources for User [%s]",
      $sMyName, $sSiebelUser );
      $iReturnCode = 0; # Failure..
    }
  }

  #-----------------------------
  # Backup siebns.dat for GTWY..
  #-----------------------------
  if ( $sServerType eq 'GTWY' &&
       Siebel::BackupSiebNs ( $sSiebelRoot ) ) {
    &logTrace ( 20048,
    "<%s> Backed up siebns.dat successfully", $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN String SiebelRoot,
                        IN String SiebelUser,
                        IN String EnvFile,
                        IN String ServerName,
                        IN String ServerType,
                        IN String SiebelEnterprise,
                        IN Integer CleanCode );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $sSiebelRoot               Absolute path to location of Siebel Server
 $sSiebelUser               UNIX user for Siebel Server
 $sEnvFile                  Absolute path to enviroment file for Siebel Server
 $sServerName               Siebel Server name if type is SRVR
 $sServerType               Type of Siebel Server
 $sSiebelEnterprise         Siebel Enterprise of this Server
 $iCleanCode                Why clean was called
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelRoot,
       $sSiebelUser,
       $sEnvFile,
       $sServerName,
       $sServerType,
       $sSiebelEnterprise,
       $iCleanCode ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanCode } ||         # Reason..
                    $iCleanCode ) =~ s/_/ /g;

  &logTrace ( 20049, "Subroutine <%s> called with:
  SiebelRoot               [%s]
  SiebelUser               [%s]
  EnvFile                  [%s]
  ServerName               [%s]
  ServerType               [%s]
  SiebelEnterprise         [%s],
  Clean reason             [%s]",

  $sMyName, $sSiebelRoot,  $sSiebelUser,  $sEnvFile,  $sServerName,
  $sServerType, $sSiebelEnterprise, $sCleanReason );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + Siebel::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $sOsdfFile, $sShmFile, $sSvcFile,
       $iTimeout, $iExitCode, @lsFiles, @lsCmdout ) = ();

  #-----------------------------
  # Select appropriate command..
  #-----------------------------
  if ( $sServerType eq 'GTWY' ) {
    $sCommand = "$sSiebelRoot/bin/stop_ns";
    $sCommandArgs = "";
  }
  else { # Siebel Server..
    $sCommand = "$sSiebelRoot/bin/stop_server";
    $sCommandArgs = "-e $sSiebelEnterprise -M $sServerName";
  }

  unless ( $hsCleanReasons { $iCleanCode } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanCode } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    &logInfo ( 20050,
    "<%s> Attempting to stop Siebel [%s] with Command [%s] Args [%s]",
    $sMyName, $sServerType, $sCommand, $sCommandArgs );
  
    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20051,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
      $iReturnCode = 0; # Failure..
    }
    elsif ( Siebel::ExecLocalCmd ( $sSiebelUser, # Ignore return code..
                                   $sEnvFile,
                                   $sCommand,
                                   $sCommandArgs,
                                   \$iExitCode,
                                   $iTimeout,
                                   \@lsCmdout ) ) {
      &logTrace ( 20052, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }

  #----------------------------------------
  # Whack any persisting Siebel processes..
  #----------------------------------------
  &logInfo ( 20053,
  "<%s> Killing processes owned by [%s]", $sMyName, $sSiebelUser );

  if ( Siebel::KillSiebelProcesses ( $sSiebelUser,
                                     $sSiebelRoot,
                                     $sServerName,
                                     $sServerType,
                                     $sSiebelEnterprise ) ) {
    &logTrace ( 20054,
    "<%s> Processes for User [%s] do not exist after shutdown",
    $sMyName, $sSiebelUser );
  }
  else {
    &logError ( 20055,
    "<%s> Could not kill all processes for User [%s]", $sMyName, $sSiebelUser );
    $iReturnCode = 0; # Failure..
  }

  #----------------------------------
  # Remove persisting IPC resources..
  #----------------------------------
  unless ( $bSharedLogin ) {
    if ( Proc::RemoveIPCResourcesForUser ( $sSiebelUser, O_IPCALL ) ) {
      &logTrace ( 20056,
      "<%s> IPC resources for User [%s] do not exist after shutdown",
      $sMyName, $sSiebelUser );
    }
    else {
      &logError ( 20057,
      "<%s> Could not remove all IPC resources for User [%s]",
      $sMyName, $sSiebelUser );
      $iReturnCode = 0; # Failure..
    }
  }

  #---------------------
  # Remove stale files..
  #---------------------
  if ( $sServerType eq 'SRVR' ) {
    #-------------------------
    # Define files to remove..
    #-------------------------
    $sOsdfFile = "$sSiebelRoot/sys/osdf.$sSiebelEnterprise.$sServerName";
    $sShmFile  = "$sSiebelRoot/admin/$sSiebelEnterprise.$sServerName.shm";
    $sSvcFile  = "$sSiebelRoot/sys/svc.siebsrvr.$sSiebelEnterprise:$sServerName";

    #-------------------------------------------
    # Identify .shm and .osdf files if present..
    #-------------------------------------------
    if ( -z $sOsdfFile ) {
      &logTrace ( 20058, "<%s> Stale OSDF file [%s] located for removal",
      $sMyName, $sOsdfFile );
      push ( @lsFiles, $sOsdfFile );
    }

    if ( -e $sShmFile ) {
      &logTrace ( 20059, "<%s> Stale SHM file [%s] located for removal",
      $sMyName, $sShmFile );
      push ( @lsFiles, $sShmFile );
    }

    #--------------------------------
    # Identify corrupt service file..
    #--------------------------------
    if ( -z $sSvcFile ) { # Zero byte file..
      &logTrace ( 20060, "<%s> Located empty service file [%s]",
      $sMyName, $sSvcFile );
      push ( @lsFiles, $sSvcFile );
    }

    #-------------------
    # Remove the files..
    #-------------------
    if ( scalar @lsFiles &&
         ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
      &logError ( 20061, "<%s> Failed to remove stale file(s) [%s]",
      $sMyName, join ( ' ', @lsFiles ) );
    }
  }
  elsif ( $sServerType eq 'GTWY' &&
          Siebel::BackupSiebNs ( $sSiebelRoot ) ) {
    &logTrace ( 20062,
    "<%s> Backed up siebns.dat successfully.", $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String SiebelRoot,
                                   IN String SiebelUser,
                                   IN String ServerName,
                                   IN String ServerType,
                                   IN String SiebelEnterprise,
                                   IN String HostName );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sSiebelRoot               Absolute path to location of Siebel Server
 $sSiebelUser               UNIX user for Siebel Server
 $sServerName               Siebel Server name if type is SRVR
 $sServerType               Type of Siebel Server
 $sSiebelEnterprise         Name of the Siebel Enterprise of this Server
 $sHostName                 Hostname/IPv4 address of server hosting this server

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelRoot,
       $sSiebelUser,
       $sServerName,
       $sServerType,
       $sSiebelEnterprise,
       $sHostName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20063, "Subroutine <%s> called:
  SiebelRoot               [%s]
  SiebelUser               [%s]
  ServerName               [%s]
  ServerType               [%s]
  SiebelEnterprise         [%s]
  HostName                 [%s]",

  $sMyName, $sSiebelRoot, $sSiebelUser, $sServerName, $sServerType,
  $sSiebelEnterprise, $sHostName);

  #------------------------------------------
  # Check if virtual hostname is up locally..
  #------------------------------------------
  unless ( VCSagentFW::RunGetProcState ( $sHostName ) ) {
    &logTrace ( 20064,
    "<%s> Virtual Host [%s] not up locally", $sMyName, $sHostName );
    return VCSDef_MonitorExit::OFFLINE;
  }

  my @liDummyPids = ();
  my $iUid = getpwnam $sSiebelUser; # Has to succeed..
  my $iReturnCode = Siebel::GetProcState ( \@liDummyPids,
                                           $sSiebelRoot,
                                           $sSiebelUser,
                                           $sServerName,
                                           $sServerType,
                                           $sSiebelEnterprise,
                                           1 ); # Fast..

  if( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20065,
    "<%s> Found all required processes", $sMyName );
  }
  else {
    &logTrace ( 20066,
    "<%s> did not find all required processes, returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;

  &logTrace ( 20067,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids
                               IN String SiebelRoot,
                               IN String SiebelUser,
                               IN String ServerName,
                               IN String ServerType,
                               IN String SiebelEnterprise,
                               IN Boolean UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sSiebelRoot               Absolute path to location of Siebel Server
 $sSiebelUser               UNIX user for Siebel Server
 $sServerName               Siebel Server name if type is SRVR
 $sServerType               Type of Siebel Server
 $sSiebelEnterprise         Name of the Siebel Enterprise of this Server
 $bUseFFLM                  Flag to indicate if fast FLM is to be attempted

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sSiebelRoot,
       $sSiebelUser,
       $sServerName,
       $sServerType,
       $sSiebelEnterprise,
       $bUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20068, "Subroutine <%s> called with:
  rliOnlinePids      [%s]
  SiebelRoot         [%s]
  SiebelUser         [%s]
  ServerName         [%s]
  ServerType         [%s]
  SiebelEnterprise   [%s] 
  UseFFLM            [%s]",

  $sMyName, $rliOnlinePids, $sSiebelRoot, $sSiebelUser,
  $sServerName, $sServerType, $sSiebelEnterprise, $bUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iFilteredProcCount, $iUseRFLM, $iFound, $iUid,
       $iPid, $sLang, $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues,
       %hiProcPids, %hsSrvrDetails, %hsCopy, %hiAppHash, %hiOth ) = ();

  #--------------------------
  # Other processes to kill..
  #--------------------------
  my @lsOthers = ( qr|\s$sSiebelRoot/| );

  $iUid = getpwnam $sSiebelUser; # Should not fail..
  @$rliOnlinePids = ();          # Reset..

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( $bUseFFLM != 2 ) { # Called for first level check..
    if ( Siebel::GetSrvrDetails ( $sServerName,
                                  $sServerType,
                                  $sSiebelRoot,
                                  $sSiebelEnterprise,
                                  \%hsSrvrDetails ) ) {
  
      &logTrace ( 20069,
      "<%s> Got details for monitoring ServerType [%s]",
      $sMyName, $sServerType );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20070,
      "<%s> Failed to get details for monitoring ServerType [%s]",
      $sMyName, $sServerType );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $bUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nSiebel process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = Siebel::FFLMStatus ( $sPidFile,
                                         $sProc,
                                         $iUid,
                                         \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
          $hiProcPids { $sKey } = $iPid;   # Store..
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20071,
      "<%s> Siebel [%s] server process details ---%s\n\n",
      $sMyName, $sServerType, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20072,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sSiebelUser, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20073,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20074,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20075,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20076,
      "<%s> Searching PID [%s] PPID [%s] UID [%s] CMD [%s]",
      $sMyName, $$rhlProc { PID }, $$rhlProc { PPID }, $$rhlProc { UID },
      $$rhlProc { COMMAND } );

      if ( $$rhlProc { UID } == $iUid ) { # Owned by SiebelUser..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $bUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( ( VCSDef_MonitorExit::OFFLINE == $iFound ) &&  # Earlier marked offline..
                   ( $$rhlProc { COMMAND } =~ /^\b$sProc\b/ ) ) { # Now found..
    
                &logTrace ( 20077,
                "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                $sProc );
                $hiProcPids { $sKey } = $iPid;        # Store..
                unshift ( @$rliOnlinePids, $iPid );   # Store..
                $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
              }
    
              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
        else { # Request for kill..
          &logTrace ( 20078,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      }

      #---------------------------------
      # Handle other related processes..
      #---------------------------------
      if ( ( ! grep { /^$iPid$/ } @$rliOnlinePids ) &&           # Not caught earlier..
           $$rhlProc { UID } == $iUid ) {                        # Owned by User
        if ( map { $$rhlProc { COMMAND } =~ /$_/ } @lsOthers ) { # Was now found..
          push ( @$rliOnlinePids, $iPid );                       # Push..
          ++ $hiAppHash { $iPid };                               # And store..

          &logTrace ( 20079,
          "<%s> Matched process PID [%s] PPID [%s] CMD [%s]",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
        }
        else {
          $hiOth { $iPid } = $$rhlProc { PPID }; # Store the PPID..
        }
      }
    }

    #----------------------------------
    # Now get child processes to kill..
    #----------------------------------
    if ( scalar keys %hiAppHash && scalar keys %hiOth ) {
      push ( @$rliOnlinePids,
      grep { defined $hiAppHash { $hiOth { $_ } } } keys %hiOth );
    }

    ( $iPid, %hiAppHash, %hiOth ) = (); # Free..
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $sServerType                              &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20080,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( Siebel::CreatePidFile ( $sPidFile,
                                     $hiProcPids { $sKey } ) ) {
          &logTrace ( 20081,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20082,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN; # Verify..

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20083,
    "<%s> Found [%s] PIDs [%s] for Siebel [%s] server.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $sServerType, $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $sServerType =~ /\S/o ) {
    &logTrace ( 20084,
    "<%s> Siebel [%s] Server process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sServerType, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  return $iReturnCode;
}

 
=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String SiebelRoot,
                          IN String SiebelUser,
                          IN String ServerName,
                          IN String ServerType,
                          IN String Sadmin,
                          IN String SadminCrPasswd,
                          IN String SiebelGWHost,
                          IN Integer SiebelGWPort,
                          IN String EnvFile,
                          IN String SiebelEnterprise,
                          IN String CompGrps,
                          IN Integer SecondLevelMonitor,
                          IN String MonitorProgram );

 Description: Performs the in-depth monitoring

 Parameters :

 $sSiebelRoot               Absolute path to location of Siebel Server
 $sSiebelUser               UNIX user for Siebel Server
 $sServerName               Siebel Server name if type is SRVR
 $sServerType               Type of Siebel Server
 $sSadmin                   Siebel level administrative login
 $sSadminCrPasswd           Crypted passwd for Siebel administrative login
 $sSiebelGWHost             Virtual Hostname hosting the Gateway
 $iSiebelGWPort             Port allocated for the Gateway
 $sEnvFile                  Absolute path to enviroment file for Siebel Server
 $sSiebelEnterprise         Siebel Enterprise of this Server
 $sCompGrps                 Comma separated list of compgrps to be monitored
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 $sMonitorProgram           Full path to external monitor program
 

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelRoot,
       $sSiebelUser,
       $sServerName,
       $sServerType,
       $sSadmin,
       $sSadminCrPasswd,
       $sSiebelGWHost,
       $iSiebelGWPort,
       $sEnvFile,
       $sSiebelEnterprise,
       $sCompGrps,
       $iSecondLevelMonitor,
       $sMonitorProgram ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20085, "Subroutine <%s> called with:
  SiebelRoot               [%s]
  SiebelUser               [%s]
  ServerName               [%s]
  ServerType               [%s]
  Sadmin                   [%s]
  SadminCrPasswd           [%s]
  SiebelGWHost             [%s]
  SiebelGWPort             [%s]
  EnvFile                  [%s]
  SiebelEnterprise         [%s]
  CompGrps                 [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]",

  $sMyName, $sSiebelRoot, $sSiebelUser, $sServerName, $sServerType,
  $sSadmin,   $sSadminCrPasswd,   $sSiebelGWHost,   $iSiebelGWPort,
  $sEnvFile, $sSiebelEnterprise, $sCompGrps,  $iSecondLevelMonitor,
  $sMonitorProgram );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         Siebel::DelayedAttributeCheck ( $sSiebelRoot,
                                         $sServerName,
                                         $sServerType,
                                         $sSadmin,
                                         $sSadminCrPasswd,
                                         $sSiebelGWHost,
                                         $iSiebelGWPort,
                                         $sEnvFile,
                                         $sCompGrps ) ) ) {
      &logTrace ( 20086,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20087,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20088,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + Siebel::GetEngineTimeout ( 1 ); # With margin..

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         Siebel::SecondLevelMonitor ( $sSiebelRoot,
                                      $sSiebelUser,
                                      $sServerName,
                                      $sServerType,
                                      $sSadmin,
                                      $sSadminCrPasswd,
                                      $sSiebelGWHost,
                                      $iSiebelGWPort,
                                      $sEnvFile,
                                      $sSiebelEnterprise,
                                      $sCompGrps,
                                      $iTillWhen ) ) ) {
      &logTrace ( 20089,
      "<%s> Siebel [%s] Server online with status [%s] after SLM",
      $sMyName, $sServerType, $iReturnCode );
    }
    else {
      &logError ( 20090,
      "<%s> Siebel [%s] Server NOT online. Status [%s]",
      $sMyName, $sServerType, $iReturnCode );
      return $iReturnCode; # Get out..
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + Siebel::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         Siebel::ExternalMonitor ( $sSiebelUser,
                                   $sEnvFile,
                                   $sExtCmd,
                                   $sExtCmdArgs,
                                   $iTillWhen ) ) ) {
      &logTrace ( 20091,
      "<%s> Siebel [%s] found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $sServerType, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20092,
      "<%s> Siebel [%s] NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sServerType, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String SiebelRoot,
                                        IN String ServerName,
                                        IN String ServerType,
                                        IN String Sadmin,
                                        IN String SadminCrPasswd,
                                        IN String SiebelGWHost,
                                        IN Integer SiebelGWPort,
                                        IN String EnvFile,
                                        IN String CompGrps );

 Description : Just In Time Validation

 Parameters :

 $sSiebelRoot               Absolute path to location of Siebel Server
 $sServerName               Siebel Server name if type is SRVR
 $sServerType               Type of Siebel Server
 $sSadmin                   Siebel level administrative login
 $sSadminCrPasswd           Crypted passwd for Siebel administrative login
 $sSiebelGWHost             Virtual Hostname hosting the Gateway
 $iSiebelGWPort             Port allocated for the Gateway
 $sEnvFile                  Absolute path to enviroment file for Siebel Server
 $sCompGrps                 Comma separated list of compgrps to be monitored
 

 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelRoot,
       $sServerName,
       $sServerType,
       $sSadmin,
       $sSadminCrPasswd,
       $sSiebelGWHost,
       $iSiebelGWPort,
       $sEnvFile,
       $sCompGrps ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20093, "Subroutine <%s> called with:
  SiebelRoot               [%s]
  ServerName               [%s]
  ServerType               [%s]
  Sadmin                   [%s]
  SadminCrPasswd           [%s]
  SiebelGWHost             [%s]
  SiebelGWPort             [%s]
  EnvFile                  [%s]
  CompGrps                 [%s]",

  $sMyName,  $sSiebelRoot, $sServerName, $sServerType, $sSadmin,
  $sSadminCrPasswd, $sSiebelGWHost,  $iSiebelGWPort,   $sEnvFile,
  $sCompGrps );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #------------------------------------------
  # Validate each of the attributes in turn..
  #------------------------------------------

  #----------------------------
  # Validate environment file..
  #----------------------------
  if ( ! -e $sEnvFile ) {
    &logError ( 20094,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-----------------------------------------
  # Validate Siebel installation directory..
  #-----------------------------------------
  if ( ! -d $sSiebelRoot ) {
    &logError ( 20095, "<%s> Server Root [%s] does not exist",
    $sMyName, $sSiebelRoot );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #------------------------------------
  # For Siebel Enterprise server only..
  #------------------------------------
  if ( $sSadmin =~ /\S/o ) {
    if ( $sServerType eq 'SRVR' ) {
      #--------------------
      # Validate CompGrps..
      #--------------------
      if ( $sCompGrps !~ /([^(,|\s)]+)/g ) {
        &logError ( 20096,
        "<%s> CompGrps cannot be NULL for Sadmin [%s]",
        $sMyName, $sSadmin );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
      }
    }
    else { # Sadmin specified for GTWY..
      &logTrace ( 20097,
      "<%s> Sadmin [%s] specified for ServerType [%s].
      Authentication enforced for Siebel GTWY Server",
      $sMyName, $sSadmin, $sServerType );
    }

    #-----------------------------
    # Validate SadminCrPasswd
    #-----------------------------
    if ( $sSadminCrPasswd !~ /\S/o ) {
      &logError ( 20098,
      "<%s> SadminCrPasswd cannot be NULL for Sadmin [%s]",
      $sMyName, $sSadmin );
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
  }
  elsif ( $sServerType eq 'SRVR' ) {
    &logError ( 20099,
    "<%s> Sadmin [%s] cannot be NULL for ServerType [%s]",
    $sMyName, $sSadmin, $sServerType );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #----------------------
  # Validate ServerName..
  #----------------------
  if ( $sServerType eq 'SRVR' &&
       $sServerName !~ /\S/o ) {
    &logError ( 20100,
    "<%s> ServerName cannot be NULL for ServerType [%s]",
    $sMyName, $sServerType );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-----------------------------------
  # Validate Siebel Gateway Server ... 
  #-----------------------------------
  if ( ! inet_aton ( $sSiebelGWHost ) ) {
    &logError ( 20101, "<%s> SiebelGWHost [%s] is invalid",
    $sMyName, $sSiebelGWHost );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String SiebelRoot,
                                     IN String SiebelUser,
                                     IN String ServerName,
                                     IN String ServerType,
                                     IN String Sadmin,
                                     IN String SadminCrPasswd,
                                     IN String SiebelGWHost,
                                     IN Integer SiebelGWPort,
                                     IN String EnvFile,
                                     IN String SiebelEnterprise,
                                     IN String CompGrps,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sSiebelRoot               Absolute path to location of Siebel Server
 $sSiebelUser               UNIX user for Siebel Server
 $sServerName               Siebel Server name if type is SRVR
 $sServerType               Type of Siebel Server
 $sSadmin                   Siebel level administrative login
 $sSadminCrPasswd           Crypted passwd for Siebel administrative login
 $sSiebelGWHost             Virtual Hostname hosting the Gateway
 $iSiebelGWPort             Port allocated for the Gateway
 $sEnvFile                  Absolute path to enviroment file for Siebel Server
 $sSiebelEnterprise         Siebel Enterprise of this Server
 $sCompGrps                 Comma separated list of compgrps to be monitored
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelRoot,
       $sSiebelUser,
       $sServerName,
       $sServerType,
       $sSadmin,
       $sSadminCrPasswd,
       $sSiebelGWHost,
       $iSiebelGWPort,
       $sEnvFile,
       $sSiebelEnterprise,
       $sCompGrps,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20103, "Subroutine <%s> called with:
  SiebelRoot               [%s]
  SiebelUser               [%s]
  ServerName               [%s]
  ServerType               [%s]
  Sadmin                   [%s]
  SadminCrPasswd           [%s]
  SiebelGWHost             [%s]
  SiebelGWPort             [%s]
  EnvFile                  [%s]
  SiebelEnterprise         [%s]
  CompGrps                 [%s]
  TillWhen                 [%s]",

  $sMyName, $sSiebelRoot, $sSiebelUser, $sServerName, $sServerType,
  $sSadmin,   $sSadminCrPasswd,   $sSiebelGWHost,   $iSiebelGWPort,
  $sEnvFile, $sSiebelEnterprise, $sCompGrps, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sOdbcFile, $sOdbcSrc, $iNow, $iTimeout, $iExitCode,
       $sCrPass, $sCommand, $sCommandArgs, $iPid, $sStatus,
       @lsCmdout, $sLang ) = ();
  my $sGWSrchStr = "started at";

  #----------
  # Gateway..
  #----------
  if ( $sServerType eq 'GTWY' ) {
    #-------------------------------------------
    # Declare SLM command and args variables..
    #-------------------------------------------
    if ( $^O eq "linux" ) {
      $sCommand = "$sSiebelRoot/bin/list_ns";
      $sCommandArgs = "";

      #-------------
      # Store LANG..
      #-------------
      $sLang = $ENV { LANG };
      $ENV { LANG } = 'C';
    } else {
      $sCommand = "$sSiebelRoot/bin/srvredit";
      $sCommandArgs = ( $sSadmin =~ /\S/ ) ?
      " -u $sSadmin -p $sSadminCrPasswd" : "";

      $sCommandArgs .= " -q -g $sSiebelGWHost"; 
      if ( $iSiebelGWPort > 0 ) {
        $sCommandArgs .= ":$iSiebelGWPort"; 
      }
      $sCommandArgs .= " -e none -z" .
      ' -c \'\\\\\$Gateway.VersionString\'';
    }

    @lsCmdout = ();

    #-------------------------------
    # Validate and execute command..
    #-------------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20104,
      "<%s> Insufficient time to continue in-depth monitoring.
      Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    elsif ( Siebel::ExecLocalCmd ( $sSiebelUser,
                                   $sEnvFile,
                                   $sCommand,
                                   $sCommandArgs,
                                   \$iExitCode,
                                   $iTimeout,
                                   \@lsCmdout,
                                   $sSadmin =~ /\S/ ?
                                   $sSadminCrPasswd : undef ) ) {
      &logTrace ( 20105,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );

      if ( $^O eq "linux" ) {
        if ( grep { /^$sGWSrchStr/o } @lsCmdout ) {
          $iReturnCode = VCSDef_MonitorExit::ONLINE;
        } else {
          $iReturnCode = VCSDef_MonitorExit::OFFLINE;
        }
      }
    }
    elsif ( ! defined $iExitCode ) {
      $lsCmdout [0] ||= 'undef'; # Set if not defined..
      $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT:
                                                 VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    else {
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }

    if ( $^O eq "linux" ) {
      #------------------------
      # Restore env variables..
      #------------------------
      defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG };
    }

    return $iReturnCode;
  }

  #----------------
  # Siebel Server..
  #----------------
  #--------------------------
  # Check if odbc.ini exits..
  #--------------------------
  $sCommand = "/bin/echo";
  $sCommandArgs = "\$ODBCINI";
  ( $iExitCode, @lsCmdout ) = (); # Re-set..

  #--------------------
  # Get time bearings..
  #--------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) {
    &logError ( 20106,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( Siebel::ExecLocalCmd ( $sSiebelUser,
                                 $sEnvFile,
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout ) ) {
    if ( ! defined ( $sOdbcFile = $lsCmdout [0] ) ) {
      &logError ( 20107,
      "<%s> ODBCINI variable not set. Impossible to connect to database.",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN; # Error..
    }
    elsif ( ! -f $sOdbcFile ) {
      &logError ( 20108,
      "<%s> ODBC init file [%s] missing. Impossible to connect to database.",
      $sMyName, $sOdbcFile );
      return VCSDef_MonitorExit::UNKNOWN; # Error..
    }
  }
  else {
    &logError ( 20109,
    "<%s> Failed to get ODBCINI environment variable", $sMyName );
    return VCSDef_MonitorExit::UNKNOWN; # Error..
  }

  #-----------------------------
  # Attempt to get ODBC source..
  #-----------------------------
  if ( open ( ODBCSRC, $sOdbcFile ) ) {
    if ( ( $sOdbcSrc ) =
         grep { s/^(siebsrvr_\w+|\w+_DSN)=.*\n$/$1/ } <ODBCSRC> ) {
      &logTrace ( 20110,
      "<%s> Using ODBC source [%s] from ODBC init file [%s]",
      $sMyName, $sOdbcSrc, $sOdbcFile );
    }
    else { # Missing..
      &logError ( 20111,
      "<%s> Cannot locate ODBC source from  ODBC init file [%s]",
      $sMyName, $sOdbcFile );
      return VCSDef_MonitorExit::UNKNOWN; # Error..
    }

    close ( ODBCSRC ) ||
    &logWarn ( 20112,
    "<%s> Close failed on ODBC init file [%s] after read",
    $sMyName, $sOdbcFile );
  }
  else {
    &logError ( 20113, "<%s> Read failed on ODBC init file [%s]",
    $sMyName, $sOdbcFile );
    return VCSDef_MonitorExit::UNKNOWN; # Error..
  }

  #-------------------------------------
  # Declare command and args variables..
  #-------------------------------------
  $sCommand = "$sSiebelRoot/bin/odbcsql";
#---- one block ----
  $sCommandArgs = "/s $sOdbcSrc << !
login $sSadmin/$sSadminCrPasswd
quit
!";
#---- one block ----

  ( $iExitCode, @lsCmdout ) = ();
  my $sSrchStr = "Outstanding transaction committed";

  #--------------------
  # Get time bearings..
  #--------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) {
    &logError ( 20114,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( Siebel::ExecLocalCmd ( $sSiebelUser,
                                 $sEnvFile,
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout,
                                 $sSadminCrPasswd ) ) {

    &logTrace ( 20115, "<%s> Looking with filter [%s] from output",
    $sMyName, $sSrchStr );

    unless ( grep { /$sSrchStr/o } @lsCmdout ) {
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..

      &logError ( 20116,
      "<%s> Connection failed with database with Cmd [%s] with args [%s]. Returning [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iReturnCode );
    }
  }
  elsif ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT:
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    &logError ( 20117,
    "<%s> Command [%s] exited with exit code [%s]. Could not connect to database",
    $sMyName, $sCommand, $iExitCode, $iReturnCode );
  }

  &logTrace ( 20118,
  "<%s> Return code set to [%s] after executing Cmd [%s] with args [%s]",
  $sMyName, $iReturnCode, $sCommand, $sCommandArgs );

  #-------------------
  # Return if failed..
  #-------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  ( $iExitCode, @lsCmdout ) = (); # Re-set..

  #----------------
  # Siebctl check..
  #----------------
  $sCommand = "$sSiebelRoot/bin/siebctl";
  $sCommandArgs = "-r $sSiebelRoot -S siebsrvr -i " .
                  "$sSiebelEnterprise:$sServerName -l -q";

  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) {
    &logError ( 20119,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( Siebel::ExecLocalCmd ( $sSiebelUser,
                                 $sEnvFile,
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout ) ) {
    #--------------------------------------
    # Check if have a PID that is running..
    #--------------------------------------
    ( $iPid ) = $lsCmdout [0] =~ m/.*pid:\s*(\d+).*/gc; # Get the pid..
    ( $sStatus ) = $lsCmdout [0] =~ m/.*\((.*)\).*/;    # Should be undef to be up..

    #---------------------------
    # Should be undef to be up..
    #---------------------------
    if ( $sStatus ) { # Oops..
      &logError ( 20120, "<%s> Found PID [%s] with status [%s]",
      $sMyName, $iPid, $sStatus );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
    else { # Okay, up..
      if ( $iPid ) {
        &logTrace ( 20121,
        "<%s> Found PID [%s] for the Siebel server",
        $sMyName, $iPid );
      }
      else { # Stopped..
        &logError ( 20122, "<%s> Siebel server not running.",
        $sMyName );
        $iReturnCode = VCSDef_MonitorExit::OFFLINE;
      }
    }
  }
  elsif ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT:
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Return, if not online..
  #------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
   return $iReturnCode;
  }

  #------------------------------
  # Define additional variables..
  #------------------------------
  my ( $sLine,
       @lsCmpFlds, @lsCmpgrpFlds, @lsCompGps,
       %hsMissCmp, %hsCmpgrps ) = ();

  ( $iExitCode, @lsCmdout ) = (); # Re-set..
  #-------------------------------------
  # Perform component level monitoring..
  #-------------------------------------
  if ( $sSadmin =~ /\S/ &&
       ( @lsCompGps = $sCompGrps =~ /([^(,|\s)]+)/g ) ) { # Defined..

    #----------------
    # Srvrmgr check..
    #----------------
    $sCommand = "$sSiebelRoot/bin/srvrmgr";

#-- one block --
  $sCommandArgs = "/u $sSadmin /e $sSiebelEnterprise " .
                  "/g $sSiebelGWHost /q /s $sServerName << !
$sSadminCrPasswd
set header false
set footer false
set delimiter +
list compgrp show CG_ALIAS, CA_RUN_STATE
set delimiter |
list comp show CC_ALIAS, CG_ALIAS, CP_RUN_STATE
quit
!";
#-- one block --
  
    #--------------------
    # Get time bearings..
    #--------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) {
      &logError ( 20123,
      "<%s> Insufficient time to continue in-depth monitoring.
      Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    elsif ( Siebel::ExecLocalCmd ( $sSiebelUser,
                                   $sEnvFile,
                                   $sCommand,
                                   $sCommandArgs,
                                   \$iExitCode,
                                   $iTimeout,
                                   \@lsCmdout,
                                   $sSadminCrPasswd ) ) {
      #---------------
      # Parse output..
      #---------------
      foreach $sLine ( @lsCmdout ) {
        if ( $sLine =~ /^Connected to 0 server/o ) {
          &logError ( 20124,
          "<%s> Could not connect to Server [%s] in Enterprise [%s].
          Cmd output [%s]",
          $sMyName, $sServerName, $sSiebelEnterprise,
          join ( chr 012, @lsCmdout ) );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
        elsif ( $sLine =~ /Could not open connection to Siebel Gateway/o ) {
          &logError ( 20125,
          "<%s> Could not connect to Gateway Server. Cmd output [%s]",
          $sMyName, join ( chr 012, @lsCmdout ) );
          $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown..
        }     
        elsif ( $sLine =~ /\|/ ) { # Check for comps..
          @lsCmpFlds = grep { s/\s//g || 1 } $sLine              # Collect..
                     =~ m/([^\|)]+)/g;                           # Shrink last field..
          if ( scalar @lsCmpFlds                        &&       # Found at least some..
               $lsCmpFlds [-1] !~ /^(Running|Online)$/o &&       # Not online or running..
               grep { /$lsCmpFlds[1]/ } @lsCompGps ) {           # From the reqd. compgrp..
  
            &logError ( 20126,
            "<%s> Found Comp [%s] in CmpGrp [%s] with State [%s]",
            $sMyName, $lsCmpFlds [0], $lsCmpFlds [1], $lsCmpFlds [-1] );
  
            push @{ $hsMissCmp { $lsCmpFlds [1] } }, $lsCmpFlds [0]; # Store..
            $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown..
          }
        }
        elsif ( $sLine =~ /\+/ ) { # Check for compgrps..
          @lsCmpgrpFlds = $sLine =~ m/([^(\s|\+)]+)/g;
  
          if ( scalar @lsCmpgrpFlds &&
               $lsCmpgrpFlds [-1] =~ /^(Running|Online)$/o ) {
            ++ $hsCmpgrps { $lsCmpgrpFlds [0] };
          }
        }
      }
  
      #------------------------------
      # Missing/defaulting compgrps..
      #------------------------------
      if ( @lsCmpgrpFlds = grep { ! $hsCmpgrps { $_ } } @lsCompGps ) {
        &logError ( 20127, "<%s> Cmpgrp(s) [%s] not Online or Running",
        $sMyName, join ( ' ', @lsCmpgrpFlds ) );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown..
      }
  
      #----------------------------------------------------
      # Report components that were not online or running..
      #----------------------------------------------------
      if ( scalar keys %hsMissCmp ) {
        foreach $sLine ( keys %hsMissCmp ) {
          &logError ( 20128,
          "<%s> CmpGrp(s) [%s] has Comp(s) [%s] not running or online",
          $sMyName, $sLine, join ( ' ', @{ $hsMissCmp { $sLine } } ) );
        }
      }
    }
    elsif ( ! defined $iExitCode ) {
      $lsCmdout [0] ||= 'undef'; # Set if not defined..
      $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT:
                                                 VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    else {
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20129,
    "<%s> Sadmin [%s] specified but CompGrps [%s] NULL.
    Component-level monitoring not requested",

    $sMyName, $sCompGrps );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : Bool ValidateSiebelParams ( IN String SiebelEnterprise );

 Description :

 Verifies if,
               i.   Exactly one resource is defined with ServerType as GTWY.
               ii.  Exactly one hostname defined to host the Gateway.
               iii. Names for Siebel servers are unique.
 Parameters :

 $sSiebelEnterprise         Name of the Siebel Enterprise of this Server

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ValidateSiebelParams {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelEnterprise ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20130, "Subroutine <%s> called with:
  SiebelEnterprise        [%s]",

  $sMyName, $sSiebelEnterprise );

  #-----------------------------------------
  # Set return code to true optimistically..
  #-----------------------------------------
  my $iReturnCode = 1;

  #------------------------------------------------------------
  # Bug in hares forces us to check whether resources have been
  # configured at all..
  #------------------------------------------------------------
  my $sCommand = $sCmdHares; # Get global..
  my $sCommandArgs = "-display -type Siebel";

  my ( $sSrvName, $sSrvType, $sEnterprise, $sGWHost, $iGWPort,
       $iExitCode, @lsCmdout, %hsSrvTypes,
       %hsGWHosts, %hiGWPorts, %hsSrvNames ) = ();

  #---------------
  # Fire command..
  #---------------
  @lsCmdout = readpipe ( "$sCommand $sCommandArgs" );
   
  #------------------
  # Get exit status..
  #------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20131,
    "<%s> No resources configured of type Siebel", $sMyName );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  $sCommandArgs = "-display -type Siebel -attribute " . 
                  "ServerName ServerType SiebelEnterprise " .
                  "SiebelGWHost SiebelGWPort";
  my $sNeededAttr = "Server(Name|Type)|Siebel(Enterprise|GW(Host|Port))";
  ( $iExitCode, @lsCmdout ) = (); # Initialise..

  #--------------------------------
  # Fire command and parse output..
  #--------------------------------
  @lsCmdout = grep {
                ! /^#/o             && # Remove header..
                ( s/\n// || 1 )     && # Chomp.. make always true..
                /\b$sNeededAttr\b/o && # And one of these..
                s/.*\s+//              # Get last field..
              } readpipe ( "$sCommand $sCommandArgs" );

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20132,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    return 0; # Error..
  }

  #-----------------------
  # Check for duplicates..
  #-----------------------
  while ( ( $sSrvName, $sSrvType, $sEnterprise, $sGWHost, $iGWPort ) =
  splice ( @lsCmdout, 0, 5 ) ) {
    &logTrace ( 20133, "<%s> Found Siebel resource with:
    ServerType           [%s]
    SiebelEnterprise     [%s]
    SiebelGWHost         [%s]
    SiebelGWPort         [%s]
    ServerName           [%s]",
    $sMyName, $sSrvType, $sEnterprise, $sGWHost, $iGWPort, $sSrvName );

    if ( ( $sSiebelEnterprise eq $sEnterprise ) ) { # When this enterpise concerns us..
      ++ $hsGWHosts  { $sGWHost };                  # Count hosts..
      ++ $hiGWPorts  { $iGWPort };                  # Count ports..
      ++ $hsSrvTypes { $sSrvType };                 # Count GTWYs..
      ++ $hsSrvNames { $sSrvName };                 # Count Server names..
    }
  }

  #--------------------------------------------
  # Check if we have exactly one GTWY defined.. 
  #--------------------------------------------
  if ( $hsSrvTypes { GTWY } &&
       ( 1 != $hsSrvTypes { GTWY } ) ) {
    &logError ( 20134,
    "<%s> Configuration error.  Exactly one GTWY server needed in Enterprise [%s]",
    $sMyName, $sSiebelEnterprise );
    $iReturnCode = 0; # Error..
  }

  #--------------------------------------------------
  # Check if we have exactly only one host for GTWY..
  #--------------------------------------------------
  unless ( 1 == keys %hsGWHosts ) {
    &logError ( 20135,
    "<%s> Configuration error.
    Multiple Gateway hostnames [%s] defined in Enterprise [%s]",
    $sMyName, join ( ' ', sort keys %hsGWHosts ), $sSiebelEnterprise );
    $iReturnCode = 0; # Error..
  }

  #--------------------------------------------------
  # Check if we have exactly only one port for GTWY..
  #--------------------------------------------------
=pod
  unless ( 1 == keys %hiGWPorts ) {
    &logError ( 20136,
    "<%s> Configuration error.
    Multiple Gateway ports [%s] defined in Enterprise [%s]",
    $sMyName, join ( ' ', sort { $a <=> $b } keys %hiGWPorts ),
    $sSiebelEnterprise );
    $iReturnCode = 0; # Error..
  }
=cut

  #-----------------------------------
  # Check for duplicate server names..
  #-----------------------------------
  foreach $sSrvName ( keys %hsSrvNames ) {
    if ( $hsSrvNames { $sSrvName } > 1 ) {
      &logError ( 20137, "<%s> Configuration error.
      ServerName [%s] appears [%s] times in Enterprise [%s]",
      $sMyName, $sSrvName, $hsSrvNames { $sSrvName }, $sSiebelEnterprise );
      $iReturnCode = 0; # Error..
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String SiebelUser,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sSiebelUser            UNIX user for Siebel Server
 $sEnvFile               Absolute path to enviroment file for Siebel Server
 $sCmd                   User defined utilty to execute
 $sCmdArgs               Command line arguments
 $iTillWhen              Time in EPOCH left for IInd level monitoring
 

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20138, "Subroutine <%s> called with:
  SiebelUser            [%s]
  EnvFile               [%s]
  Cmd                   [%s]
  CmdArgs               [%s]
  TillWhen              [%s]",

  $sMyName,
  $sSiebelUser, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT  );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20139,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20140,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout =
            $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20141,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20142, "<%s> Executing external monitor:
  Command [%s] with Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  Siebel::ExecLocalCmd ( $sSiebelUser,
                         $sEnvFile,
                         $sCmd,
                         $sCmdArgs,
                         \$iExitCode,
                         $iTimeout,
                         \@lsCmdout );

  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) { # Timeout..
    &logError ( 20143,
    "<%s> Command [%s] with args [%s] timed out",
    $sMyName, $sCmd, $sCmdArgs );
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT:
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20144,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20145,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillSiebelProcesses ( IN String SiebelUser,
                                       IN String SiebelRoot,
                                       IN String ServerName, );
                                       IN String ServerType,
                                       IN String SiebelEnterprise );

 Description  : Kills processes for requested User
 
 Parameters :

 $sSiebelUser               UNIX user for Siebel Server
 $sSiebelRoot               Absolute path to location of Siebel Server
 $sServerName               Siebel Server name if type is SRVR
 $sServerType               Type of Siebel Server
 $sSiebelEnterprise         Siebel Enterprise of this Server

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub KillSiebelProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelUser,
       $sSiebelRoot,
       $sServerName,
       $sServerType,
       $sSiebelEnterprise ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20146, "Subroutine <%s> called with:
  SiebelUser           [%s]
  SiebelRoot           [%s]
  ServerName           [%s]
  ServerType           [%s]
  SiebelEnterprise     [%s]",
 
  $sMyName,
  $sSiebelUser, $sSiebelRoot, $sServerName,
  $sServerType, $sSiebelEnterprise );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'Siebel::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    Siebel::GetProcState ( \@liGoners,
                           $sSiebelRoot,
                           $sSiebelUser,
                           $sServerName,
                           $sServerType,
                           $sSiebelEnterprise,
                           0 ); # Regluar FLM..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20147,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20148,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20149, "<%s> Could not locate any process for User [%s]",
    $sMyName, $sSiebelUser );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : float PercentCpuUtil ( IN Intger Pid );

 Description : Finds the percentage CPU utilisation of given PID

 Parameters  :

 $iPid        PID whose information is requested


 Returns : float value if successful
           undef, if not
 -------------------------------------------------------------------------

=cut

sub PercentCpuUtil {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20150, "Subroutine <%s> called with:
  PID  [%s]", $sMyName, $iPid );

  #-------------------
  # Define variables..
  #-------------------
  my ( $iPCpu, $iExitCode, $sOldUnix95, $sCommand, $sCommandArgs ) = ();
  $sCommand = "/bin/ps";
  $sCommandArgs = "-p $iPid -o pcpu=";

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20151, "<%s> Cmd [%s] does not exist",
    $sMyName, $sCommand );
    return undef; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20152, "<%s> Cmd [%s] is not executable",
    $sMyName, $sCommand );
    return undef; # Failure..
  }
  elsif ( ! defined $iPid   ||    # Not defined or..
          $iPid !~ /^\d+$/o ||    # Not a digit..
          ! kill ( 0, $iPid ) ) { # Not alive..
    &logError ( 20153, "<%s> Invalid PID [%s] passed",
    $sMyName, $iPid );
    return undef; # Failure..
  }

  #---------------
  # Support HPUX..
  #---------------
  if ( $^O eq 'hpux' ) {
     $sOldUnix95 = $ENV { UNIX95 }; # Backup..
     $ENV { UNIX95 } = 'XPG4';      # Request XPG4 behavour..
  }

  #-----------------------
  # Get %age utilization..
  #-----------------------
  ( $iPCpu = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/^\s+|\n//g;

  #----------------
  # Get exit code..
  #----------------
  $iExitCode = $? >> 8;

  #----------
  # Restore..
  #----------
  defined $sOldUnix95 ? $ENV { UNIX95 } = $sOldUnix95 : delete $ENV { UNIX95 };

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode ) {
    &logError ( 20154,
    "<%s> Cmd [%s] with Args [%s] failed with exit code [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    $iPCpu = undef; # Failure..
  }
  else {
    &logTrace ( 20155, "<%s> PID [%s]. Percent CPU utilisation [%s]",
    $sMyName, $iPid, $iPCpu );
  }

  return $iPCpu;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout ,
                                IN String HideString );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 $sHideString      String in CommandArgs to hide

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ,
       $sHideString ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20156, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  my ( $iExitCode, $sDispStr, $sCmdDeCr, $sDeCrPasswd ) = ();

  #---------------------
  # Set up for display..
  #---------------------
  $sDispStr = $sCommandArgs;

  #-------------------------
  # Get decrypted password..
  #-------------------------
  if ( defined $sHideString ) {
    #-----------
    # Validate..
    #-----------
    if ( ! -x $sCmdVcsDecrypt ) {
      &logError ( 20157,
      "<%s> Command [%s] is not executable", $sMyName, $sCmdVcsDecrypt );
      return 0; # Failure..
    }

    $sCmdDeCr = $sCmdVcsDecrypt;                          # Get global..
    ( $sDeCrPasswd ) = grep {                             # Get decrypted..
                          s/([^\w-])/\\$1/g || 1          # Escape non-words.. 
                       } readpipe ( "$sCmdDeCr $sHideString" );

    #--------------------
    # Collect exit code..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20158, "<%s> Failed to decrypt crypted string [%s]",
      $sMyName, $sHideString );
      return 0; # Failure..
    }

    $sCommandArgs =~ s/$sHideString/$sDeCrPasswd/g;
  }

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20159,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20160,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20161,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sDispStr );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20162,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sDispStr );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout,
                                                           undef,
                                                           undef,
                                                           $sDeCrPasswd );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20163,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20164,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20165,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sDispStr, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20166,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName : delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String ServerName,
                                  IN String ServerType,
                                  IN String SiebelRoot,
                                  IN String SiebelEnterprise,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sServerName               Siebel Server name if type is SRVR
 $sServerType               Type of Siebel Server
 $sSiebelRoot               Absolute path to location of Siebel Server
 $sSiebelEnterprise         Name of the Siebel Enterprise of this Server
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerName,
       $sServerType,
       $sSiebelRoot,
       $sSiebelEnterprise,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20167, "Subroutine <%s> called with:
  ServerName            [%s]
  ServerType            [%s]
  SiebelRoot            [%s]
  SiebelEnterprise      [%s]
  rhlSrvrDetails        [%s]",

  $sMyName, $sServerName, $sServerType, $sSiebelRoot,
  $sSiebelEnterprise, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sProc, $sPidFile, $sPIDDir, $sResourceType, $sResourceName ) = ();

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20168, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20169, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  if ( $sServerType eq 'SRVR' ) {
    $sPidFile = "$sPIDDir/pidfile.$sServerType";
    $sProc = "siebsvc.*?siebsrvr.*?$sSiebelEnterprise.*?$sServerName";
  }
  elsif ( $sServerType eq 'GTWY' ) {
    $sPidFile = "$sPIDDir/pidfile.$sServerType";
    $sProc = "siebsvc.*?gtwyns.*?$sSiebelRoot/sys/siebns.dat";
  }
  else {
    &logError ( 20170, "<%s> Called with in-appropriate ServerType [%s]",
    $sMyName, $sServerType );

    %$rhlSrvrDetails = (); # Truncate..
    return 0; # Failure..
  }

  &logTrace ( 20171,
  "<%s> ServerType [%s] Search Pattern [%s] PID file [%s]",
  $sMyName, $sServerType, $sProc, $sPidFile );

  #--------
  # Store..
  #--------
  @{ $$rhlSrvrDetails { $sServerType } } = ( $sPidFile, $sProc );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20172, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20173, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20174, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20175, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20176,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20177,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20178,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( Siebel::GetPidDetails ( $iPid,
                                 \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20179,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&            # Owner match..
               $$rhlProc { COMMAND } =~ /^\b$sProc\b/ ) { # Filter match..
          &logTrace ( 20180,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20181,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20182,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20183, "<%s> PID [$iPid] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20184, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20185,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  if ( $sSwitch eq "solaris" ) {
      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  } elsif ( $sSwitch eq "hpux" ) {
      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  } elsif ( $sSwitch eq "aix" ) {
      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  } elsif ( $sSwitch eq "linux" ) {
      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }

  &logTrace ( 20186, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20187,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95 : delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20188,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20189, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20190, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20191,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20192,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20193,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used

 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20194, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;

  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20195,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
  }
  else {
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    $sCommand = $sCmdHares; # Get global..
    $sCommandArgs = "-value $sResourceName $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                      =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20196,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s].",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iEngineTimeout ) { # Undefined value..
      &logTrace ( 20197, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20198, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    unless ( $iEngineTimeout ) {
      $sCommand = $sCmdHatype; # Get global..
      ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
      $sResourceType = "Siebel" if $sResourceType !~ /\w/; # Ensure..
      $sCommandArgs = "-value $sResourceType $sAttribute";
    
      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
    
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20199,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      elsif ( ! $iEngineTimeout ) { # Empty value..
        &logTrace ( 20200, "<%s> Could not get [%s] from engine",
        $sMyName, $sAttribute );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      else { # Success..
        &logTrace ( 20201, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
      }
    }
    else {
        &logTrace ( 20202, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20203,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}


=pod

 -------------------------------------------------------------------------
 Function : bool RecreateSvcFile ( IN String SiebelUser,
                                   IN String EnvFile,
                                   IN String SiebelRoot,
                                   IN String SiebelEnterprise,
                                   IN String ServerName,
                                   IN String SiebelGWHost.
                                   IN Integer SiebelGWPort.
                                   IN String ServerName,
                                   IN Integer Timeout );

 Description : Recreates a corrupted service file

 Parameters :

 $sSiebelUser               UNIX user for Siebel Server
 $sEnvFile                  Absolute path to enviroment file for Siebel Server
 $sSiebelRoot               Absolute path to location of Siebel Server
 $sSiebelEnterprise         Siebel Enterprise of this Server
 $sServerName               Siebel Server name if type is SRVR
 $sSiebelGWHost             Virtual Hostname hosting the Gateway
 $iSiebelGWPort             Port allocated for the Gateway
 $iTimeout                  Time limit to recreate the service file

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub RecreateSvcFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelUser,
       $sEnvFile,
       $sSiebelRoot,
       $sSiebelEnterprise,
       $sServerName,
       $sSiebelGWHost,
       $iSiebelGWPort,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20204, "Subroutine <%s> called with:
  SiebelUser           [%s]
  EnvFile              [%s]
  SiebelRoot           [%s]
  SiebelEnterprise     [%s]
  ServerName           [%s]
  SiebelGWHost         [%s]
  SiebelGWPort         [%s]
  Timeout              [%s]",

  $sMyName,
  $sSiebelUser, $sEnvFile, $sSiebelRoot,
  $sSiebelEnterprise, $sServerName, $sSiebelGWHost, $iSiebelGWPort, $iTimeout );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $sSvcFile, $iExitCode, @lsCmdout) = ();

  #--------------------------------------
  # Define command based on server type..
  #--------------------------------------
  if ( $sServerName =~ /\S/ ) { # SRVR
    $sSvcFile = "$sSiebelRoot/sys/svc.siebsrvr.$sSiebelEnterprise:$sServerName";
    $sCommand = "$sSiebelRoot/bin/siebctl";
    $sCommandArgs = "/S siebsrvr /i $sSiebelEnterprise:$sServerName /a " .
           "/g \"/g $sSiebelGWHost  /e $sSiebelEnterprise /s $sServerName\"";
  }
  else { # GTWY
    $sSvcFile = "$sSiebelRoot/sys/svc.gtwyns.";
    $sCommand = "$sSiebelRoot/bin/siebctl";
    $sCommandArgs = "/S gtwyns /a /g \"/f $sSiebelRoot/sys/siebns.dat";
    if ( $iSiebelGWPort > 0 ) {
      $sCommandArgs .= " /t $iSiebelGWPort";
    }
    $sCommandArgs .= "\"";
  }

  #--------------------------------
  # Return if service file exists..
  #--------------------------------
  if ( -e $sSvcFile ) {
    &logWarn ( 20205,
    "<%s> Service file [$sSvcFile] exists. Will not recreate.",
    $sMyName, $sSvcFile );
    return 0;
  }

  #-----------
  # Validate..
  #-----------
  if ( $iReturnCode =
       Siebel::ExecLocalCmd ( $sSiebelUser, # Attempt to re-create it..
                              $sEnvFile,
                              $sCommand,
                              $sCommandArgs,
                              \$iExitCode,
                              $iTimeout,
                              \@lsCmdout ) ) {
    #-----------------
    # Verify success..
    #-----------------
    if ( $lsCmdout [-1] =~ /^created at/o ) { # Success..
      &logTrace ( 20206, "<%s> Service file [%s] successfully [%s]",
      $sMyName, $sSvcFile, $lsCmdout [-1] );
    }
    else {
      &logError ( 20207,
      "<%s> Failed to recreate service file [%s]", $sMyName, $sSvcFile );
      $iReturnCode = 0; # Failure..
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool BackupSiebNs ( IN String SiebelRoot );

 Description : Backs-up siebns.dat file

 Parameters :

 $sSiebelRoot               Absolute path to location of Siebel Server

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub BackupSiebNs {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSiebelRoot ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20208, "Subroutine <%s> called with:
  SiebelRoot  [%s]",
  $sMyName, $sSiebelRoot );

  #------------------------------------
  # Derive current resource type/name..
  #------------------------------------
  my $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  my $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sBackupDir = "/var/tmp/.VRTS$sResourceType/$sResourceName/BACKUP";
  my $sSiebNSBack = "$sBackupDir/siebns.dat";
  my $sSiebNSDat = "$sSiebelRoot/sys/siebns.dat";
  
  my ( $iUid, $iGid, $sDetails ) = ();

  #--------------------------
  # Create backup directory..
  #--------------------------
  if ( -d $sBackupDir ) {
    &logTrace ( 20209,
    "<%s> Backup directory [%s] exits", $sMyName, $sBackupDir );
  }
  elsif ( ! mkpath ( $sBackupDir, 0, 0700 ) ) {
    &logWarn ( 20210,
    "<%s> Failed to create backup directory [%s]: [%s]",
    $sMyName, $sBackupDir, $! ); 
    return 0;
  }

  #--------------------------
  # Get details and proceed..
  #--------------------------
  if ( $sDetails = stat ( $sSiebNSDat ) ) {
    #--------------
    # Get details..
    #--------------
    $iUid = $sDetails->uid;
    $iGid = $sDetails->gid;

    &logTrace ( 20211,
    "<%s> Backing up last valid siebns.dat [%s] to [%s]",
    $sMyName, $sSiebNSDat, $sSiebNSBack );

    if ( open ( SIEBNS, $sSiebNSDat ) ) {
      binmode SIEBNS; # Binary read..

      #----------------------
      # Open file for write..
      #----------------------
      if ( sysopen ( BACKUP, $sSiebNSBack,
                         O_WRONLY|O_SYNC|O_CREAT|O_TRUNC, 0644 ) ) {

        binmode BACKUP;
        print BACKUP <SIEBNS>; # Fast copy..

        #---------------------
        # Close backup file..
        #---------------------
        unless ( close ( BACKUP ) ) {
          &logWarn ( 20212,
          "<%s> Failed to close [%s] after write: [%s]",
          $sMyName, $sSiebNSBack, $! );
          $iReturnCode = 0; # Failure..
        }

        #-------------------
        # Change ownership..
        #-------------------
        unless ( chown ( $iUid, $iGid, $sSiebNSBack ) ) {
          &logWarn ( 20213,
          "<%s> Failed to chown file [%s] to UID [%s] GID [%s]: [%s]",
          $sMyName, $sSiebNSBack, $iUid, $iGid, $! );
          $iReturnCode = 0; # Failure..
        }
      }
      else {
        &logWarn ( 20214,
        "<%s> Failed to open backup file [%s] for write: [%s]",
        $sMyName, $sSiebNSBack, $! );
        $iReturnCode = 0; # Failure..
      }

      #----------------------
      # Close original file..
      #----------------------
      unless ( close ( SIEBNS ) ) {
        &logWarn ( 20215,
        "<%s> Failed to close [%s] after read: [%s]",
        $sMyName, $sSiebNSDat, $! );
        $iReturnCode = 0; # Failure..
      }
    }
    else {
      &logWarn ( 20216,
      "<%s> Failed to open [%s] for read: [%s]",
      $sMyName, $sSiebNSDat, $! );
      $iReturnCode = 0; # Failure..
    }
  }
  else {
    &logWarn ( 20217,
    "<%s> Failed to get details for [%s]: [%s]",
    $sMyName, $sSiebNSDat, $! );
  }

  return $iReturnCode;
}


1; # So PERL knows its included..
